/* -*- mode: c; c-basic-offset: 4; -*- */
#include "daqc.h"
#include "daqc_internal.h"
#include "trench.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#define MAX_GROUP_LIST   4096
#define MAX_CHANNEL_LIST 262144

#define NEW_VECT(type,dim) ((type*)malloc(dim*sizeof(type)))

enum pgm_mode {
    list_group,
    chan_list,
    chan_data,
    cmd_exec
};

#define DEFAULT_DELTA 2

/************************************************************************
 *                                                                      *
 *       Dump out data                                                  *
 *                                                                      *
 ************************************************************************/
void
dump_data(daq_t* daq) {
    int i;
    for (i=0; i<daq->num_chan_request; i++) {
	chan_req_t* chan = daq->chan_req_list + i;
	if (chan->status < 0) {
	    printf("Channel: %s receive error (%i)\n", 
		   chan->name, -chan->status);	    
	    continue;
	}

	daq_data_t dtype = chan->data_type;
	size_t N = (size_t)chan->status / data_type_size(dtype);
	printf("Channel: %s  type: %s  nWords: %zi\n", 
	       chan->name, data_type_name(dtype), N);

	size_t j, k;
	switch (dtype) {
	case _16bit_integer: {
	    short* p = (short*)(daq->tb->data + chan->offset);
	    for (j=0; j<N; j+=8) {
		printf("%5zi", j);
		for (k=j; k<j+8 && k<N; ++k) 
		    printf(" %6i", (int)*p++);
		printf("\n");
	    }
	    break;
	}
	case _32bit_integer: {
	    int32_t* p = (int32_t*)(daq->tb->data + chan->offset);
	    for (j=0; j<N; j+=8) {
		printf("%5zi", j);
		for (k=j; k<j+8 && k<N; ++k) 
		    printf(" %8i", (int)*p++);
		printf("\n");
	    }
	    break;
	}
	case _32bit_float: {
	    float* p = (float*)(daq->tb->data + chan->offset);
	    for (j=0; j<N; j+=8) {
		printf("%5zi", j);
		for (k=j; k<j+8 && k<N; ++k) 
		    printf(" %8.3f", *p++);
		printf("\n");
	    }
	    break;
	}
	case _64bit_double: {
	    double* p = (double*)(daq->tb->data + chan->offset);
	    for (j=0; j<N; j+=8) {
		printf("%5zi", j);
		for (k=j; k<j+8 && k<N; ++k) 
		    printf(" %8.3f", *p++);
		printf("\n");
	    }
	    break;
	}
	default:
	    printf("Channel: %s data of unknown/unsupprted type (%i)\n", 
		   chan->name, chan->data_type);
	}
    }
}

/************************************************************************
 *                                                                      *
 *       Dump out channel status information                            *
 *                                                                      *
 ************************************************************************/
void
dump_status(daq_t* daq) {
    size_t i;
    printf("Data for %i seconds starting at GPS: %i\n", 
	   daq->tb->secs, daq->tb->gps);
    printf("%-40s %-7s %s\n", "Channel", "type", "nWords");
    for (i=0; i<daq->num_chan_request; i++) {
	chan_req_t* chan = daq->chan_req_list + i;
	daq_data_t dtype = chan->data_type;
	if (chan->status < 0) {
	    printf("%-40s %-7s: %s\n", chan->name, 
		   data_type_name(chan->data_type), 
		   daq_strerror(-chan->status));
	} else {
	    size_t N = (size_t) chan->status / data_type_size(dtype);
	    printf("%-40s %-7s %zi\n", chan->name, 
		   data_type_name(chan->data_type), N);
	}
    }
}

/************************************************************************
 *                                                                      *
 *       Main Routine                                                   *
 *                                                                      *
 ************************************************************************/
int
main(int argc, const char* argv[]) {
    enum nds_version vrsn = nds_try;
    enum pgm_mode mode   = chan_data;
    enum chantype c_type = cUnknown;

    short port_id   = DAQD_PORT;
    const char* cmd = 0;
    const char* node_id = "localhost";
    time_t gps_start=0, gps_end=0, delta=DEFAULT_DELTA;
    int verbose = 0;

    int iarg;
    int syntax=0;
    for (iarg=1; iarg<argc && *(argv[iarg]) == '-'; ++iarg) {
	if (!strcmp(argv[iarg], "-c")) {
	    mode = cmd_exec;
	    cmd  = argv[++iarg];
	} else if (!strcmp(argv[iarg], "-d")) {
	    delta = strtol(argv[++iarg], 0, 0);
	} else if (!strcmp(argv[iarg], "-e")) {
	    gps_end = strtol(argv[++iarg], 0, 0);
	} else if (!strcmp(argv[iarg], "-g")) {
	    mode = list_group;
	} else if (!strcmp(argv[iarg], "--help")) {
	    syntax = 1;
	    break;
        } else if (!strcmp(argv[iarg], "-l")) {
	    mode = chan_list;
	} else if (!strcmp(argv[iarg], "-n")) {
	    node_id = argv[++iarg];
	} else if (!strcmp(argv[iarg], "-p")) {
	    port_id = (short) strtol(argv[++iarg], 0, 0);
	} else if (!strcmp(argv[iarg], "-s")) {
	    gps_start = strtol(argv[++iarg], 0, 0);
	} else if (!strcmp(argv[iarg], "-t")) {
	    c_type = cvt_str_chantype(argv[++iarg]);
	} else if (!strcmp(argv[iarg], "-v")) {
	    verbose++;
	} else if (!strcmp(argv[iarg], "-1")) {
	    vrsn = nds_v1;
	} else if (!strcmp(argv[iarg], "-2")) {
	    vrsn = nds_v2;
	} else {
	    fprintf(stderr, "Unrecognized command: %s\n", argv[iarg]);
	    syntax = 1;
	    break;
	}
    }
    if (mode == chan_data && iarg == argc) {
	fprintf(stderr, "No channels requested!\n");
	syntax = 1;
    }
    if (syntax) {
	fprintf(stderr, "Command syntax: \n");
	fprintf(stderr, "daq_test [-g] [-l] [-n <server>] [-p <port>] [-s <start-gps>]\n");
	fprintf(stderr, "         [-e <end-gps>] [-d <delta-t>] [<channel-list>]\n");
	fprintf(stderr, "where: \n");
	fprintf(stderr, "    -d      Time stride [%i]\n", DEFAULT_DELTA);
	fprintf(stderr, "    -e      End gps time [start_gps + delta]\n");
	fprintf(stderr, "    -g      Request a list of channel groups (NDS1 only)\n");
	fprintf(stderr, "    -l      List available channels\n");
	fprintf(stderr, "    -n      Specify server IP address\n");
	fprintf(stderr, "    -p      Specify port number [%i]\n", DAQD_PORT);
	fprintf(stderr, "    -s      Specify start time (as GPS)\n");
	fprintf(stderr, "    -t      Specify default channel type.\n");
	fprintf(stderr, "    -v      Verbose printout (dump channel data)\n");
	fprintf(stderr, "    -1      NDS1 protocol\n");
	fprintf(stderr, "    -2      NDS2 protocol\n");
	fprintf(stderr, "    <channel-list> Space delimited channel list\n");
	fprintf(stderr, "If the protocol is not specified, daq_test tries\n");
	fprintf(stderr, "nds2 protocol, and if this fails, reverts to nds1\n");
	fprintf(stderr, "protocol\n");
	return 1;
    }
    daq_startup();

    /*--------------------------------------------------------------------*
     *                                                                    *
     *      Connect to the NDS server                                     *
     *                                                                    *
     *--------------------------------------------------------------------*/
    if (!port_id) port_id = DAQD_PORT;
    daq_t daq;
    int rc = daq_connect(&daq, node_id, port_id, vrsn);
    if (rc) {
	printf("Error in daq_connect: %s\n", daq_strerror(rc));
	return 1;
    }

    /*--------------------------------------------------------------------*
     *                                                                    *
     *      Execute a specific command                                    *
     *                                                                    *
     *--------------------------------------------------------------------*/
    if (mode == cmd_exec) {
	rc = daq_send(&daq, (char*)cmd);
	printf("command sent, rc=%i\n", rc);

    /*--------------------------------------------------------------------*
     *                                                                    *
     *      Get a group list in group list mode.                          *
     *                                                                    *
     *--------------------------------------------------------------------*/
    } else if (mode == list_group) {
	daq_channel_group_t* group_list = 
	    NEW_VECT(daq_channel_group_t,MAX_GROUP_LIST);
	int nGroups = 0;
	rc = daq_recv_channel_groups(&daq, group_list, 
				     MAX_GROUP_LIST, &nGroups);
	if (rc) {
	    printf("Error in daq_recv_channel_groups: %s\n", 
		   daq_strerror(rc));
	    return 2;
	}
	int i;
	printf(" Group    ID\n");
	for (i=0; i<nGroups; i++) {
	    printf("%8s %5i\n", group_list[i].name, group_list[i].group_num);
	}

    /*--------------------------------------------------------------------*
     *                                                                    *
     *      Get the channel list                                          *
     *                                                                    *
     *--------------------------------------------------------------------*/
    } else {
	/*------------------------------  Count the number of channels    */
	int nAlloc = 0;
	rc = daq_recv_channel_list(&daq, 0, 0, &nAlloc, gps_start, c_type);
	if (nAlloc > MAX_CHANNEL_LIST) nAlloc = MAX_CHANNEL_LIST;

	/*------------------------------  Read in the channel list        */
	int nChans = 0;
	daq_channel_t* channel_list = NEW_VECT(daq_channel_t, (size_t)nAlloc);
	rc = daq_recv_channel_list(&daq, channel_list, MAX_CHANNEL_LIST, 
				   &nChans, gps_start, c_type);
	if (rc) {
	    printf("Error reading channel list: %s\n", daq_strerror(rc));

	/*-------------------------------------------------------------------*
	 *                                                                   *
	 *      Dump the channel list in list mode                           *
	 *                                                                   *
	 *-------------------------------------------------------------------*/ 
	} else if (mode == chan_list) {
	    int i;
	    printf("Number of channels received = %i\n", nChans);
	    printf("Channel                  Rate  chan_type\n");
	    for (i=0; i<nChans; i++) {
		printf("%-40s %5g %8s %9s\n", channel_list[i].name, 
		       channel_list[i].rate, 
		       cvt_chantype_str(channel_list[i].type),
		       data_type_name(channel_list[i].data_type)
		       );
	    }
	} else {

	/*-------------------------------------------------------------------*
	 *                                                                   *
	 *     Get and dump the channel data for the specified channels      *
	 *                                                                   *
	 *-------------------------------------------------------------------*/
	    int i;
	    for (i=iarg; i<argc; ++i) {
		struct trench_struct tch;
		trench_init(&tch);
		trench_parse(&tch, argv[i]);
		int j, chanok = 0;
		daq_channel_t chan;
		if (daq.nds_versn == nds_v1) {
		    for (j=0; j<nChans; ++j) {
			if (!trench_cmp_base(&tch, channel_list[j].name)) {
			    daq_init_channel(&chan, argv[i], cUnknown,
					     channel_list[j].rate,
					     channel_list[j].data_type);
			    chanok = 1;
			    break;
			}
		    }
		} else {
		    for (j=0; j<nChans; ++j) {
			if (!strcmp(tch.str, channel_list[j].name)) {
			    daq_init_channel(&chan, argv[i],
					     channel_list[j].type,
					     channel_list[j].rate,
					     channel_list[j].data_type);
			    chanok = 1;
			    break;
			}
		    }
		}

		if (chanok) {
		    daq_request_channel_from_chanlist(&daq, &chan);
		} else {
		    printf("Channel: %s not in list... Ignored.\n", argv[i]);
		}
	    }

	/*-------------------------------------------------------------------*
	 *                                                                   *
	 *     Get and dump the channel data for the specified channels      *
	 *                                                                   *
	 *-------------------------------------------------------------------*/
	    if (!gps_end) {
		gps_end = gps_start + (time_t) delta;
	    } else if (gps_end < gps_start) {
		gps_end += gps_start;
	    }
	    if (gps_end < gps_start + delta) {
		delta = gps_end - gps_start;
	    }

	    rc = daq_request_data(&daq, gps_start, gps_end, delta);
	    if (rc) {
		printf("Error in daq_request_data: %s\n", daq_strerror(rc));
	    } else {
		time_t gps;
		for (gps=gps_start; gps<gps_end; gps+=delta) {
		    int rc = daq_recv_next(&daq);
		    if (!rc) {
			if (verbose) dump_data(&daq);
			else         dump_status(&daq);
		    } else {
			printf("Error in daq_recv_next: %s\n", 
			       daq_strerror(rc));
			break;
		    }
		}
	    }
	}
    }

    /*--------------------------------------------------------------------*
     *                                                                    *
     *     All done, go away...                                           *
     *                                                                    *
     *--------------------------------------------------------------------*/
    daq_disconnect(&daq);
    daq_recv_shutdown(&daq);
    return 0;
}
